##############################################################################
#
#    Recurrence Rules for Odoo
#    Copyright © 2014, 2018, 2022 XCG Consulting (https://www.xcg-consulting.fr/)
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

import ast
import datetime
from collections.abc import Mapping
from typing import Any

import dateutil.rrule
from odoo import api, fields, models
from weekdays.helpers import BusinessCalendar

# The dateutil.rrule.values are not stored as is because YEARLY is 0 and Odoo thinks
# that the field is not set in that case.
# Dictionary that match stored values to dateutil.rrule values.
_freq_dict = {
    "YEARLY": dateutil.rrule.YEARLY,
    "MONTHLY": dateutil.rrule.MONTHLY,
    "WEEKLY": dateutil.rrule.WEEKLY,
    "DAILY": dateutil.rrule.DAILY,
    "HOURLY": dateutil.rrule.HOURLY,
    "MINUTELY": dateutil.rrule.MINUTELY,
    "SECONDLY": dateutil.rrule.SECONDLY,
}


class RRuleDay(models.Model):
    """this model is used by all implementers who need to define days that will be
    passed to our rrule engine.

    A good example of this would be a timesheet application that would need to define
    the start of the week and the day-off days per people."""

    _name = "rrules.rrule_day"
    _description = "Recurrence rule - Day"

    name = fields.Char(string="Name", required=True, translate=True)

    weekday = fields.Integer(string="Weekday", required=True)


class RRule(models.Model):
    """Model representing a recurrence rule"""

    _name = "rrules.rrule"
    _description = "Recurrence Rule"

    name = fields.Char(string="Name", size=64, required=True, translate=True)

    # From here, the field are in the same order as dateutil.rrule doc

    freq = fields.Selection(
        selection=[
            ("YEARLY", "Yearly"),
            ("MONTHLY", "Monthly"),
            ("WEEKLY", "Weekly"),
            ("DAILY", "Daily"),
            ("HOURLY", "Hourly"),
            ("MINUTELY", "Minute"),
            ("SECONDLY", "Secondly"),
        ],
        string="Frequency",
        help="Frequency of the recurrence rule.",
        required=True,
    )
    # cache: no reason to let it be set by user
    interval = fields.Integer(
        string="Interval",
        default=1,
        help="The interval between each freq iteration. For example, when"
        " using YEARLY, an interval of 2 means once every two years, but"
        " with HOURLY, it means once every two hours. The default interval"
        " is 1.",
    )
    # dtstart: not to be set by rrule, only when evaluating one
    # wkst is that useful? can be added if needed
    # count this should not be set, we’ll use dtstart/until
    # until: not set by rrule, only when evaluating one on a period
    bysetpos = fields.Char(
        string="By set pos",
        size=64,
        help="If given, it must be either an integer, or a sequence of"
        " integers, positive or negative. Each given integer will specify"
        " an occurrence number, corresponding to the nth occurrence of the"
        " rule inside the frequency period. For example, a bysetpos of -1"
        " if combined with a MONTHLY frequency, and a byweekday of (MO,"
        " TU, WE, TH, FR), will result in the last work day of every month"
        ".",
    )
    bymonth = fields.Char(
        string="By month",
        size=64,
        help="If given, it must be either an integer, or a sequence"
        " of integers, meaning the months to apply the recurrence to.",
    )
    bymonthday = fields.Char(
        string="By month day",
        size=64,
        help="If given, it must be either an integer, or a sequence"
        " of integers, meaning the month days to apply the recurrence to.",
    )
    byyearday = fields.Char(
        string="By year day",
        size=64,
        help="If given, it must be either an integer, or a sequence"
        " of integers, meaning the year days to apply the recurrence to.",
    )
    byweekno = fields.Char(
        string="By week no",
        size=64,
        help="If given, it must be either an integer, or a sequence of"
        " integers, meaning the week numbers to apply the recurrence to."
        " Week numbers have the meaning described in ISO8601, that is,"
        " the first week of the year is that containing at least four days"
        " of the new year.",
    )
    byweekday = fields.Char(
        string="By week day",
        size=64,
        help="If given, it must be an integer (0 == MO), a sequence of"
        " integers. When given, these variables will define the weekdays"
        " where the recurrence will be applied. It's also possible to use"
        " an argument n for the weekday instances, which will mean the nth"
        " occurrence of this weekday in the period. For example, with"
        " MONTHLY, or with YEARLY and BYMONTH, using FR(+1) in byweekday"
        " will specify the first Friday of the month where the recurrence"
        " happens. Notice that in the RFC documentation, this is specified"
        " as BYDAY, but was renamed to avoid the ambiguity of that keyword"
        ".",
    )
    # TODO add byhour?
    # TODO add byminute?
    # TODO add bysecond?
    byeaster = fields.Char(
        string="By easter",
        size=64,
        help="If given, it must be either an integer, or a sequence of"
        " integers, positive or negative. Each integer will define an"
        " offset from the Easter Sunday. Passing the offset 0 to byeaster"
        " will yield the Easter Sunday itself. This is occidental Easter,"
        " not orthodox Easter.",
    )

    def get_rrule(self, args=None) -> dateutil.rrule.rrule:
        """Return a dateutil rrule from a rrule model"""

        self.ensure_one()

        return dateutil.rrule.rrule(**self.get_rrule_kwargs(args=args))

    def get_rrule_kwargs(self, args: Mapping[str, Any] = None) -> dict[str, Any]:
        """
        :return dict: kwargs for a rrule creation.
        """

        self.ensure_one()

        if args is None:
            args = {}

        rrule_kwargs = {
            "freq": _freq_dict[self.freq],
            "interval": self.interval,
        }
        if "date_from" in args:
            rrule_kwargs["dtstart"] = args["date_from"]
        if "date_end" in args:
            rrule_kwargs["until"] = args["date_end"]
        if self.bysetpos:
            rrule_kwargs["bysetpos"] = ast.literal_eval(self.bysetpos)
        if self.bymonth:
            rrule_kwargs["bymonth"] = ast.literal_eval(self.bymonth)
        if self.bymonthday:
            rrule_kwargs["bymonthday"] = ast.literal_eval(self.bymonthday)
        if self.byyearday:
            rrule_kwargs["byyearday"] = ast.literal_eval(self.byyearday)
        if self.byweekno:
            rrule_kwargs["byweekno"] = ast.literal_eval(self.byweekno)
        if self.byweekday:
            rrule_kwargs["byweekday"] = ast.literal_eval(self.byweekday)
        if self.byeaster:
            rrule_kwargs["byeaster"] = ast.literal_eval(self.byeaster)

        return rrule_kwargs

    @api.model
    def get_rrule_today(
        self, dayoff: dict = None, base_date: datetime.datetime = None
    ) -> dateutil.rrule.rrule:
        """Get an RRule for the specified base date (defaults to today).

        :param dayoff: Argument to handle holidays. TODO To document.
        :param base_date: The date to build an RRule for. Optional; defaults to today.
        """
        # TODO Document the "dayoff" kwarg in the above comment.

        if base_date is None:
            today = datetime.date.today()
            base_date = datetime.datetime(today.year, today.month, today.day)
        args = {"date_from": base_date, "date_end": base_date}

        rrule_kwargs = self.get_rrule_kwargs(args=args)

        if dayoff:
            # any week starting day is ok there
            business_calendar = BusinessCalendar(
                0,
                dayoff["weekoff"],
                [rrule.get_rrule_kwargs(args=args) for rrule in dayoff["holidays"]],
            )

            # evaluate dayoff for the period/freq including today
            if (
                rrule_kwargs["freq"] == dateutil.rrule.MONTHLY
                or rrule_kwargs["freq"] == dateutil.rrule.WEEKLY
            ):
                # In the case of weekly, all the dates are not necessary
                rrule_kwargs["bymonthday"] = [
                    datetime_.day
                    for datetime_ in business_calendar.get_month_business_days(
                        base_date.year, base_date.month
                    )
                ]
            else:
                raise NotImplementedError

        # get back rrules
        return dateutil.rrule.rrule(**rrule_kwargs)
