##############################################################################
#
#    Recurrence Rules for Odoo
#    Copyright © 2017, 2018, 2022 XCG Consulting <https://xcg-consulting.fr/>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################
import datetime

from odoo import fields, tests


class Test(tests.TransactionCase):
    """Test for rrules."""

    def test_get_rrule(self):
        """Test around rrules.rrule::get_rrule."""

        # The new year recurrence rule.
        rrule = self.env.ref("rrules.new_year").get_rrule()

        # The first iteration is on the next January 1st.
        first_year = datetime.date.today().year + 1

        # Next iterations should all be on January 1st.
        for i in range(3):
            next_date = rrule[i].date()
            self.assertEqual(next_date, datetime.date(first_year + i, 1, 1))

    def test_get_rruleset(self):
        """Test around rrules.rruleset::get_rruleset."""

        # Ask for rules around the 14th of July (French holiday). In 2017, that
        # was on a Friday.
        rruleset = self.env.ref("rrules.france_working_days").get_rruleset(
            args={"date_from": datetime.date(2017, 7, 14)}
        )

        # Next open dates start on Monday 17th. Check a whole week.
        self.assertEqual(rruleset[0].date(), datetime.date(2017, 7, 17))  # Mon
        self.assertEqual(rruleset[1].date(), datetime.date(2017, 7, 18))  # Tue
        self.assertEqual(rruleset[2].date(), datetime.date(2017, 7, 19))  # Wed
        self.assertEqual(rruleset[3].date(), datetime.date(2017, 7, 20))  # Thu
        self.assertEqual(rruleset[4].date(), datetime.date(2017, 7, 21))  # Fri
        self.assertEqual(rruleset[5].date(), datetime.date(2017, 7, 24))  # Mon

    def test_rrule_preview(self):
        """Check the recurrence rule preview dialog box."""

        preview_dlg = (
            self.env["rrules.rrule_preview"]
            .with_context(
                active_id=self.env.ref("rrules.new_year").id,
                active_model="rrules.rrule",
            )
            .create({})
        )

        # The first iteration is on the next January 1st.
        first_year = datetime.date.today().year + 1

        # These are the same as those in the first test (test_0000_get_rrule).
        for i in range(0, 7):
            next_date = fields.Datetime.to_datetime(
                getattr(preview_dlg, "next%d" % i),
            ).date()
            self.assertEqual(next_date, datetime.date(first_year + i, 1, 1))
